/**
 * @file   copy.c
 * @brief  Third-party copy implementation.
 * @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
 */
#include <apr_buckets.h>
#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/io.h>
#include <httpd.h>
#include <http_log.h>
#include <http_protocol.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <util_filter.h>

#include "mod_lcgdm_disk.h"
#include "../mod_dav/mod_dav.h"
#include "../client/htext.h"
#include "../shared/utils.h"

/** Structure used to pass data along callbacks. */
typedef struct {
  apr_bucket_brigade *brigade;
  ap_filter_t        *output;
  request_rec        *request;
} dav_disk_copy_data;

/* Custom I/O handler */
typedef struct {
  off_t                  fsize;
  dmlite_context*        context;
  const dmlite_any_dict* extra;
} dav_disk_open_params;



off_t dav_disk_fsize(const char *path, void *udata)
{
  dav_disk_open_params* params = udata;
  return params->fsize;
}



void* dav_disk_fopen(const char *path, const char *mode, void *udata)
{
  dav_disk_open_params* params = udata;
  
  /* In fact, we only want read mode! */
  if (strchr(mode, 'w') != NULL || strchr(mode, '+') != NULL)
    return NULL;  
  
  return dmlite_fopen(params->context, path, O_RDONLY, params->extra);
}



int dav_disk_fclose(void *fd)
{
  return dmlite_fclose(fd);
}



size_t dav_disk_fread(void *buffer, size_t size, size_t n, void *fd)
{
  return dmlite_fread(fd, buffer, size * n);
}



size_t dav_disk_fwrite(const void* buffer, size_t size, size_t n, void *fd)
{
  return 0;
}



int dav_disk_fseek(void *fd, long offset, int whence)
{
  return dmlite_fseek(fd, offset, whence);
}



long dav_disk_ftell(void* fd)
{
  return dmlite_ftell(fd);
}



int dav_disk_feof(void* fd)
{
  return dmlite_feof(fd);
}



struct htext_io_handler dav_disk_io_handler = {
  dav_disk_fsize,
  dav_disk_fopen,
  dav_disk_fclose,
  dav_disk_fread,
  dav_disk_fwrite,
  dav_disk_fseek,
  dav_disk_ftell,
  dav_disk_feof
};


/** @brief  Return a pointer to a malloc'd string holding a URL-encoded
 *          (RFC 1738) version of "in" . Only A-Z a-z 0-9 . _ - are passed
 *          through unmodified.
 * @details DN's processed by GRSThttpUrlEncode can be used as valid
 *          Unix filenames, assuming they do not exceed restrictions on
 *          filename length. Adapted from GridSite.
 */
static char *GRSThttpUrlEncode(apr_pool_t *pool, const char *in)
{
  char       *out, *q;
  const char *p;
  int         outlen;

  outlen = 3 * strlen(in) + 1;
  out = apr_pcalloc(pool, outlen);

  p = in;
  q = out;

  while (*p != '\0') {
   if (isalnum(*p) || (*p == '.') || (*p == '_') || (*p == '-')) {
    *q = *p;
    ++q;
   }
   else if (*p == ' ') {
     /* Spaces are replaced by + by GridSite */
     snprintf(q, outlen, "%%2B");
     q = &q[3];
   }
   else {
     snprintf(q, outlen, "%%%2X", *p);
     q = &q[3];
   }
   ++p;
  }

  *q = '\0';
  return out;
}




/**
 * Called by htext when there is logging output
 * @param handle The handle that triggers the call
 * @param type   Type of logging data
 * @param msg    The message itself
 * @param size   The size of the message (it might not be NULL terminated)
 * @param ud     User defined data
 */
static void dav_disk_copy_log(htext_handle *handle, HTEXT_LOG_TYPE type,
                              const char *msg, size_t size, void *ud)
{
  dav_disk_copy_data *ddc = (dav_disk_copy_data*)ud;

  switch (type) {
   case HTEXT_LOG:
     ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ddc->request,
                   "%s", msg);
     break;
    default:
      break;
 }
}



/* Exposed function. */
dav_error *dav_disk_remote_copy(const dav_resource *src, const char *remote,
                                int depth)
{
  apr_uri_t            remote_uri;
  dav_error           *error = NULL;
  apr_bucket          *bkt;
  const char          *error_string;
  char                *uproxy;
  struct stat          fstat;
  htext_handle        *handle;
  int                  wait, status;
  dav_disk_copy_data   ddc;
  struct timespec      sleep_time = {0, 5000};
  dav_disk_open_params open_params;
  
  /* Parse URI */
  apr_uri_parse(src->pool, remote, &remote_uri);

  /* Local copies are not accepted */
  if (apr_strnatcasecmp(remote_uri.hostname,
                        src->info->request->server->server_hostname) == 0)
  {
    return dav_shared_new_error(src->info->request, NULL, HTTP_METHOD_NOT_ALLOWED,
                               "Local copies not allowed");
  }

  /* Check permissions */
  if (!(src->info->d_conf->flags & DAV_DISK_REMOTE_COPY)) {
    return dav_shared_new_error(src->info->request, NULL, HTTP_METHOD_NOT_ALLOWED,
                                "This server does not allow remote copies");
  }

  /* Get delegation proxy */
  if (!src->info->delegation) {
    return dav_shared_new_error(src->info->request, NULL, HTTP_BAD_REQUEST,
                                "A delegation id is needed for remote copies");
  }

  uproxy = apr_pstrcat(src->pool,
                       src->info->d_conf->proxy_cache,
                       "/%3A",
                       GRSThttpUrlEncode(src->pool, src->info->user->client_name),
                       "/", src->info->delegation, "/userproxy.pem",
                       NULL);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                "Using delegation proxy '%s'", uproxy);

  /* Sanity check */
  if (stat(uproxy, &fstat) != 0)
    return dav_shared_new_error(src->info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                "Could not stat %s", uproxy);
  if (!S_ISREG(fstat.st_mode))
    return dav_shared_new_error(src->info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                               "%s is not a file", uproxy);

   /* Setup the handle */
  if (!(handle = htext_init()))
    return dav_shared_new_error(src->info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                "Not enough memory");

  open_params.fsize   = src->info->fsize;
  open_params.context = src->info->ctx;
  open_params.extra   = src->info->loc.chunks[0].url.query;

  htext_setopt(handle, HTEXTOP_SOURCEURL,      src->info->loc.chunks[0].url.path);
  htext_setopt(handle, HTEXTOP_IO_HANDLER,     &dav_disk_io_handler);
  htext_setopt(handle, HTEXTOP_IO_HANDLER_DATA,&open_params);
  htext_setopt(handle, HTEXTOP_DESTINATIONURL, remote);

  htext_setopt(handle, HTEXTOP_USERCERTIFICATE, uproxy);
  htext_setopt(handle, HTEXTOP_USERPRIVKEY,     uproxy);
  htext_setopt(handle, HTEXTOP_CAFILE,          uproxy);

  htext_setopt(handle, HTEXTOP_CAPATH,     "/etc/grid-security/certificates");
  htext_setopt(handle, HTEXTOP_VERIFYPEER, 1);

  ddc.output  = src->info->request->output_filters;
  ddc.brigade = apr_brigade_create(src->pool, ddc.output->c->bucket_alloc);
  ddc.request = src->info->request;
  htext_setopt(handle, HTEXTOP_LOGCALLBACK,      dav_disk_copy_log);
  htext_setopt(handle, HTEXTOP_LOGCALLBACK_DATA, &ddc);
  htext_setopt(handle, HTEXTOP_VERBOSITY,        1);

  /* Run */
  if (htext_perform(handle) != 0) {
    error= dav_shared_new_error(src->info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                "Could not perform the action: %s",
                                htext_error_string(handle));
    htext_destroy(handle);
    return error;
  }

  /* Wait */
  src->info->request->status = 0;
  bkt = apr_bucket_flush_create(ddc.output->c->bucket_alloc);
  do {
    size_t *total, *done, n, i;
    size_t  globalDone, globalTotal;

    status = htext_status(handle);
    switch (status) {
      case HTEXTS_SUCCEEDED: case HTEXTS_FAILED:
      case HTEXTS_ABORTED:
        wait = 0;
        break;
      default:
        /* In the first go we need to set the reply */
        if (src->info->request->status == 0) {
          src->info->request->status = HTTP_ACCEPTED;
          ap_set_content_type(src->info->request, "text/plain");
        }
        /* Print progress */
        htext_progress(handle, &n, &total, &done);
        globalDone = globalTotal = 0;
        for (i = 0; i < n; ++i) {
          globalDone  += done[i];
          globalTotal += total[i];
        }
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output,
                           "%lu/%lu\n",
                           (unsigned long)globalDone, (unsigned long)globalTotal);
        ap_fflush(ddc.output, ddc.brigade);
        /* Sleep a while */
        wait = 1;
        nanosleep(&sleep_time, NULL);
    }
  } while (wait);

  /* Feedback */
  error_string = htext_error_string(handle);
  switch (status) {
    case HTEXTS_SUCCEEDED:
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, src->info->request,
                    "Remote copy finished successfully (%d): %s => %s",
                    htext_http_code(handle), src->uri, remote);
      /* 200 is OK, but 201 fits better */
      if (src->info->request->status == 0)
        src->info->request->status = HTTP_CREATED;
      apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output, "Success\n");
      break;
    case HTEXTS_FAILED:
      if (src->info->request->status == 0)
        error = dav_shared_new_error(src->info->request, NULL,
                                     htext_http_code(handle),
                                     "Failed: Remote copy failed with status code %d: %s\n",
                                     htext_http_code(handle),
                                     error_string?error_string:"");
      else
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output,
                           "Failed: Remote copy failed with status code %d: %s\n",
                           htext_http_code(handle),
                           error_string?error_string:"");
      break;
    case HTEXTS_ABORTED:
      if (src->info->request->status == 0)
        error = dav_shared_new_error(src->info->request, NULL,
                                     HTTP_INTERNAL_SERVER_ERROR,
                                     "Aborted");
      else
        apr_brigade_printf(ddc.brigade, ap_filter_flush, ddc.output, "Aborted\n");
      break;
  }

  /* Finish */
  htext_destroy(handle);

  if (!error) {
    bkt = apr_bucket_eos_create(ddc.output->c->bucket_alloc);
    APR_BRIGADE_INSERT_TAIL(ddc.brigade, bkt);
    if (ap_pass_brigade(ddc.output, ddc.brigade) != APR_SUCCESS)
      error = dav_shared_new_error(src->info->request, NULL,
                                   HTTP_INTERNAL_SERVER_ERROR,
                                   "Could not write EOS to filter.");
  }

  return error;
}
