/**
 * @file   mod_lcgdm_ns.h
 * @brief  Configurations and hooks.
 * @author Alejandro Álvarez Ayllón <aalvarez.cern.ch>
 */
#ifndef MOD_LCGDM_NS_H_
#define MOD_LCGDM_NS_H_

#include <apr_tables.h>
#include <dmlite/c/dmlite.h>
#include <dmlite/c/inode.h>
#include <httpd.h>
#include <http_config.h>
#include "../mod_dav/mod_dav.h"
#include "../shared/security.h"

#define DAV_NS_WRITE             0x01
#define DAV_NS_NOAUTHN           0x02

/** Node types. */
typedef enum {
  DAV_NS_NODE_HEAD,
  DAV_NS_NODE_LFC
} node_type;
typedef unsigned char dir_flags;

/* Data that needs to be available outside. */
extern module AP_MODULE_DECLARE_DATA lcgdm_ns_module;
extern const dav_hooks_repository    dav_ns_hooks_repository;
extern const dav_liveprop_group      dav_ns_liveprop_group;
extern const dav_hooks_db            dav_ns_hooks_db;

/** Data structure for server configuration. */
struct dav_ns_server_conf
{
  dmlite_manager *manager;
  node_type       type;
};
typedef struct dav_ns_server_conf dav_ns_server_conf;

/** Data structure for directory configuration. */
struct dav_ns_dir_conf
{
  const char *redir_scheme;
  char       *anon_user, *anon_group;
  dir_flags   flags;

  unsigned    max_replicas;

  apr_array_header_t *trusted_dns;
};
typedef struct dav_ns_dir_conf dav_ns_dir_conf;


/**
 * @brief This structure handles the data the module needs to keep track of
 *        a resource coming from a DPM service typedef done in mod_dav.h.
 */
struct dav_resource_private {
  request_rec          *request;
  dav_ns_server_conf   *s_conf;
  dav_ns_dir_conf      *d_conf;

  dmlite_context       *ctx;          /**< dmlite context. */
  dmlite_credentials   *user;

  const char           *sfn;
  const char           *redirect;     /**< Allocated only when needed. */
  dmlite_xstat          stat;

  const char           *space_token; /**< Space token. */

  int64_t               r_id;         /**< The requested replica ID in the referrer host. */
  char                **replicas;     /**< The requested replicas. */
  int64_t              *replicas_ids; /**< Corresponding IDs. */
  int                   n_replicas;   /**< The number of replicas passed. */

  const char           *forbidden_str; /**< As passed in the query. */
  const char           *notfound_str; /**< As passed in the query. */
  
  char                  metalink /**< != 0 if the user requested the metalink */;
};

/**
 * Insert all supported liveprops into the output.
 * @param r        The request.
 * @param resource The resource being processed.
 * @param what     What to insert.
 * @param phdr     Where to insert.
 */
void dav_ns_insert_all_liveprops(request_rec *r, const dav_resource *resource,
                                 dav_prop_insert what, apr_text_header *phdr);

/**
 * Find the property ID so the value can be retrieved later.
 * @param resource The resource being processed.
 * @param ns_uri   The property namespace (e.g. DAV:)
 * @param name     The property name      (e.g. getcontentlength)
 * @param hooks    Where to put the liveprop hooks.
 * @return         The property ID, or 0 if not found.
 */
int dav_ns_find_liveprop(const dav_resource *resource, const char *ns_uri,
                         const char *name, const dav_hooks_liveprop **hooks);

/**
 * Get the location of a file, and set the redirect if needed.
 * @param info Input/output parameter. sfn must be set. location will be set.
 * @param pool Pool for memory allocation.
 * @return     NULL on success.
 */
dav_error *dav_ns_get_location(dav_resource_private *info, apr_pool_t *pool);

#endif
