/**
 * @file   repository.c
 * @brief  Main part of the code: resource creating, directory listing...
 * @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
 */
#include <apr_file_io.h>
#include <apr_strings.h>
#include <ctype.h>
#include <dmlite/c/dmlite.h>
#include <dmlite/c/catalog.h>
#include <dmlite/c/pool.h>
#include <dmlite/c/utils.h>
#include <httpd.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <inttypes.h>

#include "delivery.h"
#include "mime.h"
#include "mod_lcgdm_ns.h"
#include "../shared/config.h"
#include "../shared/utils.h"

/* Used for internal recursivity */
#define DAV_WALKTYPE_POSTFIX 0x8000
#define DAV_CALLTYPE_POSTFIX 1000

/* Advanced declarations */
static dav_error *dav_ns_walk(const dav_walk_params *params, int depth,
                               dav_response **response);

/** Structure used for the walking */
struct dav_ns_walker_context {
  const dav_walk_params *params;

  dav_walk_resource walk_resource;
 
  dav_buffer    sfn;
  dav_buffer    uri;
};
typedef struct dav_ns_walker_context dav_ns_walker_context;

/** Structure for streams (dummy, since only empty creates will be handled) */
struct dav_stream {
};

/**
 * Get the location of the file
 * @param info Input/output parameter. sfn must be set. location will be set.
 * @param pool Pool for memory allocation
 * @return NULL on success
 */
dav_error *dav_ns_get_location(dav_resource_private *info,
                               apr_pool_t *pool)
{
  dmlite_replica *replicas;
  unsigned        n_replicas, i;
  char           *referrer, *p;

  /* If there is a rejected list, this is a fallback */
  if ((info->s_conf->type == DAV_NS_NODE_LFC) &&
      (info->forbidden_str != NULL || info->notfound_str != NULL)) {
    int64_t *rejected_ids, id;
    int      n_forbidden = 0, n_notfound = 0, n_total = 0;

    /* Count first to allocate */
    if (info->forbidden_str != NULL) {
      n_forbidden = 1;
      for (i = 0; info->forbidden_str[i] != '\0'; ++i) {
        if (info->forbidden_str[i] == ',')
          ++n_forbidden;
      }
    }
    if (info->notfound_str != NULL) {
      n_notfound = 1;
      for (i = 0; info->notfound_str[i] != '\0'; ++i) {
        if (info->notfound_str[i] == ',')
          ++n_notfound;
      }
    }

    /* Build rejected list */
    n_total      = n_forbidden + n_notfound;
    rejected_ids = apr_pcalloc(pool, sizeof(int64_t) * n_total);
    i            = 0;

    p = (char*)info->forbidden_str;
    while (p != NULL && *p != '\0') {
      errno = 0;
      id    = strtol(p, &p, 0);
      if (errno == 0)
        rejected_ids[i++] = id;
      if (*p != '\0')
        ++p;
    }

    p = (char*)info->notfound_str;
    while (p != NULL && *p != '\0') {
      errno = 0;
      id    = strtol(p, &p, 0);
      if (errno == 0)
        rejected_ids[i++] = id;
      if (*p != '\0')
        ++p;
    }

    dmlite_any* any = dmlite_any_new_long_array(n_total, rejected_ids);    
    i = dmlite_set(info->ctx, "ExcludeReplicas", any);
    dmlite_any_free(any);
    if (i != 0) {
      return dav_shared_new_error(info->request, info->ctx, 0,
                                  "Error on fall-back method");
    }
  }

  /* Logic here depends on the node type */
  dmlite_location *location;
  switch (info->s_conf->type) {
    case DAV_NS_NODE_HEAD:
      /* Ask for the best one */
      location = dmlite_get(info->ctx, info->sfn);
      switch(dmlite_errno(info->ctx)) {
        case 0:
          break;
        case EACCES:
          info->redirect = dav_shared_build_aggregation_url(info->request->pool,
                                                            info->n_replicas,
                                                            (const char**)info->replicas,
                                                            info->replicas_ids,
                                                            info->forbidden_str,
                                                            info->notfound_str,
                                                            info->r_id, -1);
          if (info->redirect != NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0,
                          info->request,
                          "Access forbidden for %s, forwarded to %s",
                          info->sfn, info->redirect);
            return NULL;
          }
        default:
          return dav_shared_new_error(info->request, info->ctx, 0, NULL);
      }
      
      /* Destination URL */
      info->redirect = dav_shared_build_url(pool, &location->chunks[0].url,
                                            info->d_conf->redir_scheme);
      
      dmlite_location_free(location);
      break;
      
    /* LFC wants the whole list */
    case DAV_NS_NODE_LFC:
      if (dmlite_getreplicas(info->ctx, info->sfn, &n_replicas, &replicas) != 0)
        return dav_shared_new_error(info->request, info->ctx, 0, NULL);

      /* Initialize */
      info->n_replicas    = n_replicas + 1;
      info->forbidden_str = NULL;
      info->notfound_str  = NULL;

      /* Build the referrer */
      if (info->request->server->port != 0) {
        referrer = apr_psprintf(pool, "%s://%s:%u%s",
                                info->d_conf->redir_scheme,
                                info->request->hostname,
                                info->request->server->port,
                                info->request->uri);
      }
      else {
        referrer = apr_psprintf(pool, "%s://%s%s",
                                info->d_conf->redir_scheme,
                                info->request->hostname,
                                info->request->uri);
      }

      /* Build the replica list, skipping the first */
      info->replicas     = apr_pcalloc(pool, sizeof(char*) * info->n_replicas);
      info->replicas_ids = apr_pcalloc(pool, sizeof(int64_t) * info->n_replicas);
      for (i = 0; i < n_replicas && i < info->d_conf->max_replicas; ++i) {
        dmlite_url* url       = dmlite_parse_url(replicas[i].rfn);
        info->replicas[i]     = dav_shared_build_url(pool, url, info->d_conf->redir_scheme);
        info->replicas_ids[i] = replicas[i].replicaid;
        dmlite_url_free(url);
      }

      /* Append ourselves */
      info->replicas[i]     = referrer;
      info->replicas_ids[i] = -1;

      /* Build the definitive URL */
      info->redirect = dav_shared_build_aggregation_url(info->request->pool,
                                                        info->n_replicas,
                                                        (const char**)info->replicas,
                                                        info->replicas_ids,
                                                        info->forbidden_str,
                                                        info->notfound_str,
                                                        -1, -1);

      /* Clean up */
      dmlite_replicas_free(n_replicas, replicas);

      break;
    default:
      return dav_shared_new_error(info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                  "Invalid node type configured");
  }

  return NULL;
}

/**
 * Return the parent path of "path"
 * @param pool Where to allocate space
 * @param path The path
 * @return     The parent path
 */
static char *dav_ns_dirname(apr_pool_t *pool, const char *path)
{
  char *parent;
  int   i;

  parent = apr_pstrcat(pool, path, NULL);

  /* We skip the last character so if path is a directory, we are fine */
  for(i = strlen(parent) - 2; i >= 0; --i) {
    if (parent[i] == '/') {
      parent[i + 1] = '\0'; /* Keep the slash */
      break;
    }
  }

  return parent;
}

/**
 * This function is used internally to create new resources.
 * @param r          The request to associate.
 * @param sfn        The SFN to link to the resource.
 * @param uri        The URI (without the query!)
 * @param user       The associated user. If NULL, will be queried and initialized.
 * @param resource   Where to put it.
 * @return           NULL on success.
 */
static dav_error *dav_ns_internal_get_resource(request_rec          *r,
                                               const char           *sfn,
                                               dmlite_credentials   *user,
                                               dav_resource        **resource)
{
  dav_resource_private *info;
  int                   exists, i;
  const char           *p;
  apr_table_t          *query;
  char                  replica_key[10];
  unsigned              nargs;

  /* Initialize info */
  info = apr_pcalloc(r->pool, sizeof(dav_resource_private));
  info->stat.extra = dmlite_any_dict_new();
  apr_pool_pre_cleanup_register(r->pool, info->stat.extra,
                                (apr_status_t(*)(void*))dmlite_any_dict_free);

  info->request  = r;
  info->sfn = sfn;

  info->s_conf = ap_get_module_config(r->server->module_config,  &lcgdm_ns_module);
  info->d_conf = ap_get_module_config(r->per_dir_config,         &lcgdm_ns_module);
  
  /* User info */
  if (user) {
    info->user = user;
  }
  else if (!(info->d_conf->flags & DAV_NS_NOAUTHN)) {
    if (!(info->user = dav_shared_get_user_credentials(r, info->ctx,
                                                       info->d_conf->anon_user,
                                                       info->d_conf->anon_group,
                                                       info->d_conf->trusted_dns)))
      return dav_shared_new_error(r, NULL, HTTP_FORBIDDEN,
                                  "Can not authenticate the user");
  }
  else {
    info->user = NULL;
  }

  /* Instantiate the catalog, if not in the connection notes */
  info->ctx = (dmlite_context*)apr_table_get(r->connection->notes, "dav_ns_ctx");
  if (info->ctx == NULL) {
    info->ctx = dmlite_context_new(info->s_conf->manager);
    if (info->ctx == NULL) {
      return dav_shared_new_error(r, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                  "Could not instantiate a context: %s",
                                  dmlite_manager_error(info->s_conf->manager));
    }

    apr_pool_pre_cleanup_register(r->connection->pool, info->ctx,
                                  (apr_status_t(*)(void*))dmlite_context_free);
    
    apr_table_setn(r->connection->notes, "dav_ns_ctx", (char*)info->ctx);
    
    if (info->user && dmlite_setcredentials(info->ctx, info->user) != 0)
        return dav_shared_new_error(r, info->ctx, HTTP_FORBIDDEN,
                                    "Could not set credentials");
  }

  /* Extract parameters from the query */
  query = dav_shared_parse_query(r->pool, r->parsed_uri.query, &nargs);

#ifdef WITH_METALINK
  /* Explicit */
  info->metalink      = (apr_table_get(query, "metalink") != NULL);
  /* Through headers maybe? (i.e. aria2c) */
  if (!info->metalink && dav_shared_request_accepts(r, "application/metalink+xml")) {
    info->metalink = 1;
  }
#else
  info->metalink      = 0;
#endif
  info->space_token   = apr_table_get(query, "spacetoken");
  info->forbidden_str = apr_table_get(query, "forbidden");
  info->notfound_str  = apr_table_get(query, "notfound");
  if ((p = apr_table_get(query, "rid")))
    info->r_id = atol(p);
  else
    info->r_id = -1;

  /* Count in advance how many replicas */
  info->n_replicas = 0;
  snprintf(replica_key, sizeof(replica_key), "r%d", info->n_replicas + 1);
  while(apr_table_get(query, replica_key) != NULL) {
    ++info->n_replicas;
    snprintf(replica_key, sizeof(replica_key), "r%d", info->n_replicas + 1);
  }

  /* Allocate space */
  info->replicas     = apr_pcalloc(r->pool, sizeof(char*)   * info->n_replicas);
  info->replicas_ids = apr_pcalloc(r->pool, sizeof(int64_t) * info->n_replicas);

  /* Go through again :/ */
  for (i = 0; i < info->n_replicas; ++i) {
    const char *field_value;

    snprintf(replica_key, sizeof(replica_key), "r%d", i + 1);
    field_value = apr_table_get(query, replica_key);

    info->replicas[i] = apr_pcalloc(r->pool, sizeof(char) * PATH_MAX);

    if (sscanf(field_value, "%"PRId64",%s",
                            &info->replicas_ids[i], info->replicas[i]) != 2) {
      return dav_shared_new_error(r, NULL, HTTP_BAD_REQUEST, "Replica parameters malformed");
    }
  }

  /* Stat */
  switch (dmlite_statx(info->ctx, info->sfn, &(info->stat))) {
    /* We are good */
    case 0:
      exists = 1;
      break;
    /* Access denied */
    case EACCES:
      /* If this is not a HEAD or a GET (both are M_GET), just fail,
       * but keep going otherwise */
      if (r->method_number != M_GET)
        return dav_shared_new_error(r, info->ctx, 0,
                                    "Access forbidden for %s on %s", info->sfn, r->method);
    case ENOENT:
      exists = 0;
      /* When creating a file, break here */
      if (r->method_number == M_PUT || r->method_number == M_MKCOL ||
          r->method_number == M_MOVE)
        break;
      /* Now, fall-back to the next replica in the list, if any */
      info->redirect = dav_shared_build_aggregation_url(info->request->pool,
                                                        info->n_replicas,
                                                        (const char**)info->replicas,
                                                        info->replicas_ids,
                                                        info->forbidden_str,
                                                        info->notfound_str,
                                                        -1, info->r_id);
      if (info->redirect != NULL) {
        apr_table_set(r->headers_out, "Location", info->redirect);
        return dav_new_error(r->pool, HTTP_MOVED_TEMPORARILY, 0,
                             info->redirect);
      }
      /* We can't :(, so go to default */
    default:
      return dav_shared_new_error(r, info->ctx, 0, "Can not stat %s", info->sfn);
  }
  
  /* Create and return resource */
  *resource = apr_pcalloc(r->pool, sizeof(dav_resource));
  (*resource)->type   = DAV_RESOURCE_TYPE_REGULAR;
  (*resource)->exists = exists;
  (*resource)->collection = ((info->stat.stat.st_mode & S_IFDIR) != 0);
  (*resource)->uri    = sfn;
  (*resource)->info   = info;
  (*resource)->hooks  = &dav_ns_hooks_repository;
  (*resource)->pool   = r->pool;

  if (exists)
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Resource for %s (%s) found (dir=%d)",
                  (*resource)->uri, (*resource)->info->sfn,
                  (*resource)->collection);
  else
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "NULL resource for %s (%s) created",
                  (*resource)->uri, (*resource)->info->sfn);

  return NULL;
}


/**
 * Generate the resource used for later operations
 * First security checkings can be done here (like user mapping)
 * @param r                The request coming from Apache
 * @param root_dir         The root dir of the request (as configured in
 *                         Apache with Location)
 * @param label            Used with version control. Not relevant.
 * @param use_checked_in   Used with version control. Not relevant.
 * @param resource         Where to put the generated resource
 * @return                 NULL on success. A dav_error instance if not.
 */
static dav_error *dav_ns_get_resource(request_rec *r,
                                      const char *root_dir,
                                      const char *label,
                                      int use_checked_in,
                                      dav_resource **resource)
{
  char                *sfn;
  dav_error           *err;
  int                  len;

  /* Compose path */
  len  = strlen(root_dir);
  if (root_dir[len - 1] == '/') {
    char *tmp = apr_pstrdup(r->pool, root_dir);
    --len;
    tmp[len] = '\0';
    root_dir = tmp;
  }

  sfn = r->parsed_uri.path ? r->parsed_uri.path : "";

  /* Create resource */
  err = dav_ns_internal_get_resource(r, sfn, NULL, resource);
  if (err)
    return err;

  /* Directories MUST finish with '/', so redirect if not */
  len = strlen(sfn);
  if ((*resource)->collection && (len == 0 || sfn[len - 1] != '/')) {
    (*resource)->info->redirect = apr_pstrcat(r->pool,
                                              ap_escape_uri(r->pool, sfn),
                                              "/",
                                              r->args ? "?" : "",
                                              r->args ? r->args : "",
                                              NULL);
    (*resource)->uri = apr_pstrcat(r->pool, (*resource)->uri, "/", NULL);
    (*resource)->info->sfn = apr_pstrcat(r->pool,
                                         (*resource)->info->sfn, "/", NULL);
  }

  /* We are done here */
  return NULL;
}

/**
 * Get the parent of a resource (if any)
 * @param resource        The resource to process
 * @param parent_resource The parent, if any
 * @return                NULL on success
 */
static dav_error *dav_ns_get_parent(const dav_resource *resource,
                                     dav_resource **parent_resource)
{
  char        *sfn;
  request_rec *dup_rec;

  /* The parent of the root is itself (kinda) */
  if (strcmp(resource->info->sfn, "/") == 0) {
    *parent_resource = (dav_resource*)resource;
    return NULL;
  }

  /* Strip last element of the path */
  sfn = dav_ns_dirname(resource->pool, resource->info->sfn);

  /* Duplicate request, removing query from the parsed URI.
   * This is because on PUT requests with query string, it will remain
   * in the parent, making get_resource to behave like it does not exist :-(
   */
  dup_rec  = apr_pcalloc(resource->pool, sizeof(request_rec));
  *dup_rec = *resource->info->request;
  dup_rec->parsed_uri.query = NULL;

  /* Return whatever the internal says */
  return dav_ns_internal_get_resource(dup_rec,
                                      sfn, resource->info->user,
                                      parent_resource);
}

/**
 * Compare two resources to check if they are the same
 * @param res1 Resource 1
 * @param res2 Resource 2
 * @return     1 if they are the same
 */
static int dav_ns_is_same(const dav_resource *res1, const dav_resource *res2)
{
  const char *destination;
  apr_uri_t  uri1, uri2;

  /* Parse resource 1 */
  apr_uri_parse(res1->pool, res1->uri, &uri1);

  /* Parse resource 2 */
  destination = apr_table_get(res2->info->request->headers_in, "destination");
  if (destination)
    apr_uri_parse(res2->pool, destination, &uri2);
  else
    apr_uri_parse(res2->pool, res2->uri, &uri2);

  /* Host may not appear */
  if (!uri1.hostname)
    uri1.hostname = res1->info->request->server->server_hostname;
  if (!uri2.hostname)
    uri2.hostname = res2->info->request->server->server_hostname;

  /* Compare */
  return strcmp(uri1.hostname, uri2.hostname) == 0 &&
         strcmp(uri1.path, uri2.path) == 0;
}

/**
 * Check if a given resource is the parent of another one
 * @param res1 Parent to check
 * @param res2 Child
 * @return     1 if res1 is the parent of res2
 */
static int dav_ns_is_parent(const dav_resource *res1, const dav_resource *res2)
{
  char *res2_parent_sfn;

  res2_parent_sfn = dav_ns_dirname(res2->pool, res2->info->sfn);

  return strcmp(res1->info->sfn, res2_parent_sfn) == 0;
}

/**
 * Open a stream. Used for PUT
 * @param resource The resource to open
 * @param mode     The mode (basically, trucate or random, which we don't support)
 * @param stream   Where to put the created structure
 * @return         NULL on success
 */
static dav_error *dav_ns_open_stream(const dav_resource *resource,
                                      dav_stream_mode mode,
                                      dav_stream **stream)
{
  dav_resource_private     *info;
  unsigned                  has_range;
  const char               *range, *length;
  int                       e;

  info = resource->info;
  
  /* Must be writable (if we are here, they want to write) */
  if (!(info->d_conf->flags & DAV_NS_WRITE))
    return dav_shared_new_error(info->request, NULL, HTTP_FORBIDDEN,
                                "Configured as read-only endpoint (%s)",
                                resource->uri);

  /* If content-length is 0, call create instead of put! */
  length = apr_table_get(info->request->headers_in, "content-length");
  if (length != NULL) {
    size_t clen = atol(length);
    if (clen == 0) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    info->request,
                    "PUT with content-length 0. Creating ns entry.");

      if (dmlite_create(info->ctx, info->sfn, 0644) != 0)
        return dav_shared_new_error(info->request, info->ctx, 0,
                                    "Could not create empty file %s", info->sfn);

      *stream = (dav_stream*)calloc(1, sizeof(dav_stream));

      return NULL;
    }
  }

  /* NS alone doesn't support PUTs with content! */
  if (info->s_conf->type == DAV_NS_NODE_LFC)
    return dav_shared_new_error(info->request, NULL, HTTP_NOT_IMPLEMENTED,
                                "LFC does not support PUTs");

  /*  Range header? */
  range = apr_table_get(info->request->headers_in, "content-range");
  has_range = (range != NULL);
  if (has_range) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                  info->request,
                  "Range: %s", range);
  }

  /* If token not set, call dm_put and see where do we have to go */
  if (info->space_token != NULL) {
    dmlite_any* any = dmlite_any_new_string(info->space_token);
    e = dmlite_set(info->ctx, "UserSpaceTokenDescription", any);
    dmlite_any_free(any);
    if (e)
      return dav_shared_new_error(info->request, info->ctx, 0,
                                 "Could not set the space token %s", info->space_token);
  }
  
  dmlite_location *location; 
  location = dmlite_put(info->ctx, info->sfn);
  switch (dmlite_errno(info->ctx)) {
    case 0:
      break;
    case EINVAL:
      return dav_shared_new_error(info->request, info->ctx, HTTP_BAD_REQUEST,
                                  "Can not get the space token %s",
                                  info->space_token);
    default:
      return dav_shared_new_error(info->request, info->ctx, 0, "Can not put %s",
                                  info->sfn);
  }      

  /* Redirect */
  info->redirect = dav_shared_build_url(resource->pool, &location->chunks[0].url,
                                        info->d_conf->redir_scheme);

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                info->request,
                "PUT request to be done in %s",
                info->redirect);

  apr_table_set(info->request->headers_out,
                "Location", info->redirect);
  
  dmlite_location_free(location);
  
  /* Standard says 301/302, but some clients will retry in that case with a GET.
   * 307 is unambiguous */
  return dav_new_error(resource->pool, HTTP_TEMPORARY_REDIRECT, 0,
                       info->redirect);
}

/**
 * Close a stream
 * @param stream The stream to close
 * @param commit Commit the change
 * @return       NULL on success
 */
static dav_error *dav_ns_close_stream(dav_stream *stream, int commit)
{
  /* Nothing to do here */
  return NULL;
}

/**
 * Write to a stream
 * @param stream  The stream where to write
 * @param buf     The data to write
 * @param bufsize How much data there is in the buffer
 * @return        NULL on success
 */
static dav_error *dav_ns_write_stream(dav_stream *stream, const void *buf,
                                      apr_size_t bufsize)
{
  /* We never enter here (either redirected, either empty PUT's) */
  return NULL;
}

/**
 * Moves to an specific position within a stream
 * @param stream       The stream to move
 * @param abs_position Where to move
 * @return             NULL on success
 */
static dav_error *dav_ns_seek_stream(dav_stream *stream, apr_off_t abs_position)
{
  /* Nothing to do here */
  return NULL;
}

/**
 * Sets the HTTP headers, like the size of the content, the modification date...
 * @param r        The request
 * @param resource The resource generated previously
 * @return         NULL on success
 */
static dav_error *dav_ns_set_headers(request_rec *r, const dav_resource *resource)
{
  dav_resource_private *info;
  dav_error            *err;
  const char           *etag;
  char                  buffer[1024];

  /* Null resource */
  if(!resource->exists)
    return NULL;

  info = resource->info;

  /* NS never serve files! Only metadata. */
  if (!resource->collection && !resource->info->metalink) {
    if ((err = dav_ns_get_location(info, r->pool)))
      return err;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Resource %s located at %s",
                   info->sfn, info->redirect);

    apr_table_setn(r->headers_out, "Location", info->redirect);
#ifdef WITH_METALINK
    snprintf(buffer, sizeof(buffer),
             "<%s://%s%s?metalink>; rel=describedby; type=\"application/metalink+xml\"",
             info->d_conf->redir_scheme, info->request->hostname, info->request->uri);
    apr_table_set(r->err_headers_out, "Link", buffer);
#endif
    return dav_new_error(r->pool, HTTP_MOVED_TEMPORARILY, 0, info->redirect);
  }

  /* Need to redirect? (Missing trailing /) */
  if (info->redirect) {
    apr_table_setn(r->headers_out, "Location", info->redirect);
    return dav_new_error(r->pool, HTTP_MOVED_PERMANENTLY, 0, info->redirect);
  }

  /* Content-type */
  if (resource->info->metalink) {
    ap_set_content_type(r, "application/metalink+xml");
    snprintf(buffer, sizeof(buffer),
             "attachment; filename=\"%s.metalink\"", info->stat.name);
    apr_table_set(r->headers_out, "Content-Disposition", buffer);
  }
  else if (resource->collection) {
    ap_set_content_type(r, "text/html");
  }

  /* Nice thing is, Apache takes care of implementing this */
  apr_table_setn(r->headers_out, "Accept-Ranges", "bytes");

  /* ETag */
  etag = resource->hooks->getetag(resource);
  apr_table_setn(r->headers_out, "ETag", etag);

  /* NS saves the time in seconds since epoch, while Apache expects microseconds */
  ap_update_mtime(r, info->stat.stat.st_mtime * 1000000);
  ap_set_last_modified(r);
    
  return NULL;
}

/**
 * This is the second most important function, as it delivers the content of the
 * file (or the directory if not DAV)
 * @param resource The resource generated previously
 * @param output   Here is where the output must be written
 * @return         NULL on success
 */
static dav_error *dav_ns_deliver(const dav_resource *resource,
                                 ap_filter_t *output)
{
  apr_bucket_brigade   *bb;
  apr_bucket           *bkt;
  dav_error            *err;

  bb   = apr_brigade_create(resource->pool, output->c->bucket_alloc);
  
  if (resource->collection)
    err = dav_ns_deliver_collection(resource, resource->info->request->output_filters, bb);
  else if (resource->info->metalink) {
    err = dav_ns_deliver_metalink(resource, resource->info->request->output_filters, bb);
  }
  else
    err = dav_new_error(resource->pool, HTTP_INTERNAL_SERVER_ERROR, 0,
                        "NS should not be trying to deliver files!");
  
  if (err != NULL)
    return err;

  /* Close and flush the output */
  bkt = apr_bucket_eos_create(output->c->bucket_alloc);
  APR_BRIGADE_INSERT_TAIL(bb, bkt);
  if (ap_pass_brigade(resource->info->request->output_filters, bb) != APR_SUCCESS)
    return dav_shared_new_error(resource->info->request, NULL, HTTP_INTERNAL_SERVER_ERROR,
                                "Could not write EOS to filter.");

  /* All OK */
  return NULL;
}

/**
 * Create a new collection (i.e. directory)
 * @param resource The resource to be created
 * @return         NULL on success
 */
static dav_error *dav_ns_create_collection(dav_resource *resource)
{
  /* Must be writable */
  if (!(resource->info->d_conf->flags & DAV_NS_WRITE))
    return dav_shared_new_error(resource->info->request, NULL, HTTP_FORBIDDEN,
                             "Configured as read-only endpoint (%s)",
                             resource->uri);

  /* Must not exist */
  if (resource->exists)
    return dav_shared_new_error(resource->info->request, 0, HTTP_CONFLICT,
                             "The file already exists (%s)",
                             resource->info->sfn);

  /* Create */
  switch (dmlite_mkdir(resource->info->ctx, resource->info->sfn, 0755)) {
    case 0:
      break;
    case ENOENT:
      return dav_shared_new_error(resource->info->request, resource->info->ctx,
                                  HTTP_CONFLICT,
                                  "Could not create the directory %s",
                                  resource->info->sfn);
    default:
      return dav_shared_new_error(resource->info->request, resource->info->ctx,
                                  0,
                                  "Could not create the directory %s",
                                  resource->info->sfn);
  }

  resource->exists     = 1;
  resource->collection = 1;
  if (dmlite_statx(resource->info->ctx, resource->info->sfn,
                   &(resource->info->stat)) != 0) {
    /* It has just been created, we shouldn't enter here, but just in case */
    return dav_shared_new_error(resource->info->request, resource->info->ctx, 0,
                             "dm_xstat failed just after the creation of %s",
                              resource->info->sfn);
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                resource->info->request,
               "%s created", resource->info->sfn);

  return NULL;
}

/**
 * Copy a resource to an non-existing one
 * Only allowed for remote copies
 * @param src       The source. It must exist
 * @param dst       The destination. Must not exist
 * @param depth     Copy in depth? (Collections)
 * @param response  The response to the copy process
 * @return          NULL on success
 */
static dav_error *dav_ns_copy(const dav_resource *src, dav_resource *dst,
                               int depth, dav_response **response)
{
  return dav_shared_new_error(src->info->request, NULL, HTTP_NOT_IMPLEMENTED,
                           "COPY not implemented");
}

/**
 * Move a resource
 * @param src       The source. It must exist
 * @param dst       The destination. It must not exist
 * @param response  The response to the move process
 * @return          NULL on success
 */
static dav_error *dav_ns_move(dav_resource *src, dav_resource *dst,
                               dav_response **response)
{
  /* Must be writable */
  if (!(src->info->d_conf->flags & DAV_NS_WRITE))
    return dav_shared_new_error(dst->info->request, 0, HTTP_FORBIDDEN,
                             "Configured as read-only endpoint (%s)",
                             dst->uri);

  /* Change the name, that's all */
  if (dmlite_rename(src->info->ctx, src->info->sfn, dst->info->sfn) != 0) {
    return dav_shared_new_error(dst->info->request, src->info->ctx, 0,
                                "Could not move from %s to %s",
                                src->info->sfn, dst->info->sfn);
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                src->info->request,
                "Moved %s to %s", src->info->sfn, dst->info->sfn);

  *response = NULL;
  return NULL;
}

/**
 * Remove a resource from DPM, either file or directory
 * NOT recursive, resource has to be a leaf, or this will fail
 * @param resource The resource to remove
 * @return         NULL on success
 */
static dav_error *dav_ns_internal_remove(const dav_resource *resource)
{
  int error = 0;

  if (resource->collection) {
    /* Remove a directory is straigh-forward */
    if (dmlite_rmdir(resource->info->ctx, resource->info->sfn) != 0) {
      error = 1;
    }
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    resource->info->request,
                    "Directory %s removed", resource->info->sfn);
    }
  }
  else {
    /* A file is not much more difficult :-) */
    if (dmlite_unlink(resource->info->ctx, resource->info->sfn) != 0) {
      error = 1;
    }
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    resource->info->request,
                    "File %s removed", resource->info->sfn);
    }
  }

  if (error) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0,
                  resource->info->request,
                  "Could not remove %s (%s)",
                  resource->info->sfn,
                  dmlite_error(resource->info->ctx));
    return dav_shared_new_error(resource->info->request, resource->info->ctx, 0,
                             "Could not remove %s", resource->info->sfn);
  }
  
  return NULL;
}

/**
 * Callback function for remove walking
 * @param wres      The walking resource
 * @param calltype  Calltype
 * @return          NULL on success
 */
static dav_error *dav_ns_remove_callback(dav_walk_resource *wres, int calltype)
{
  /* Only remove elements that exist, and empty directories */
  if (wres->resource->exists &&
      (!wres->resource->collection || calltype == DAV_CALLTYPE_POSTFIX)) {

    dav_error *err;

    err = dav_ns_internal_remove(wres->resource);

    /* We add to the multistatus, not as return */
    if (err)
      dav_shared_add_response(wres, err);
  }
  return NULL;
}

/**
 * Remove a resource
 * @param resource The resource to remove
 * @param response The response to the request
 * @return         NULL on success
 */
static dav_error *dav_ns_remove(dav_resource *resource,
                                 dav_response **response)
{
  dav_error *err;

  /* Be careful: mod_dav will try to remove the file when copying, if source
   * and dest happen to have the same path (even if the host is different).
   * Just ignore those requests.
   */
  if (resource->info->request->method_number == M_COPY) {
    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0,
                  resource->info->request,
                  "Ignoring remove requests on COPY");
    return NULL;
  }

  /* Must be writable */
  if (!(resource->info->d_conf->flags & DAV_NS_WRITE))
    return dav_shared_new_error(resource->info->request, 0, HTTP_FORBIDDEN,
                             "Configured as read-only endpoint (%s)",
                             resource->uri);

  
  if (resource->collection) {
    /* Recursive if collection */
    dav_walk_params  params = { 0 };
    dav_response    *multi_status;

    params.walk_type = DAV_WALKTYPE_POSTFIX;
    params.func      = dav_ns_remove_callback;
    params.pool      = resource->pool;
    params.root      = resource;

    if ((err = dav_ns_walk(&params, DAV_INFINITY, &multi_status)) != NULL)
      return err;

    if ((*response = multi_status) != NULL) {
      /* Multistatus */
      return dav_new_error(resource->pool, HTTP_MULTI_STATUS, 0,
                           "Errors on multistatus");
    }

    resource->exists = 0;
    resource->collection = 0;

    return NULL;
  }
  else {
    /* File is pretty easy */
    err = dav_ns_internal_remove(resource);
    if (err)
      return err;
    resource->exists     = 0;
    resource->collection = 0;
    return NULL;
  }
}

/**
 * Recursive function to walk through the filesystem
 * @param ctx The context
 * @param depth The maximum depth
 * @return NULL on success
 */
static dav_error *dav_ns_walker(dav_ns_walker_context *ctx, int depth)
{
  dav_error          *err = NULL;
  void               *dir;
  dmlite_xstat       *entry;
  apr_pool_t         *pool;
  const dav_resource *this_res;
  dav_resource        reusable;
  apr_pool_t         *subpool;

  pool = ctx->params->pool;

  /* Save for later */
  this_res = ctx->walk_resource.resource;

  /* Call the function with the current root */
  err = (*ctx->params->func)(&(ctx->walk_resource),
                               ctx->walk_resource.resource->collection ?
                                 DAV_CALLTYPE_COLLECTION: DAV_CALLTYPE_MEMBER);
  if (err)
    return err;
  if (depth == 0 || !ctx->walk_resource.resource->collection)
    return NULL;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
              ctx->walk_resource.resource->info->request,
              "Descending into %s", ctx->sfn.buf);

  /* Open the directory */
  if ((dir = dmlite_opendir(this_res->info->ctx, ctx->sfn.buf)) == NULL) {
    dav_shared_add_response(&ctx->walk_resource,
                            dav_shared_new_error(this_res->info->request,
                                                 this_res->info->ctx, 0,
                                                 "Could not open the directory %s",
                                                 ctx->sfn.buf));
    return NULL;
  }

  /* Create the subpool */
  apr_pool_create(&subpool, ctx->walk_resource.pool);

  /* Initialize the reusable resource */
  reusable = *this_res;
  reusable.info = NULL;
  reusable.uri  = NULL;
  reusable.info = apr_pcalloc(reusable.pool, sizeof(dav_resource_private));
  reusable.info->request = this_res->info->request;
  reusable.info->user    = this_res->info->user;
  reusable.info->ctx     = this_res->info->ctx;
  reusable.info->s_conf  = this_res->info->s_conf;
  reusable.info->d_conf  = this_res->info->d_conf;
  reusable.pool          = subpool;

  ctx->walk_resource.resource = &reusable;

  /* Scan the directory */
  while((entry = dmlite_readdirx(this_res->info->ctx, dir))) {
    apr_size_t    uri_prev_len;
    apr_size_t    path_prev_len;
    int len;

    apr_pool_clear(subpool);

    /* Append the file to path and URI */
    len = strlen(entry->name);
    dav_buffer_place_mem(pool, &ctx->uri, entry->name, len + 1, 1);
    dav_buffer_place_mem(pool, &ctx->sfn, entry->name, len + 1, 1);

    reusable.info->sfn = ctx->sfn.buf;
    
    /* dav_buffer_place_mem do NOT adjust len */
    uri_prev_len  = ctx->uri.cur_len;
    path_prev_len = ctx->sfn.cur_len;
    ctx->sfn.cur_len  += len;
    ctx->uri.cur_len  += len;

    if (entry->stat.st_mode & S_IFDIR) {
      ctx->sfn.cur_len++;
      ctx->uri.cur_len++;

      ctx->uri.buf[ctx->uri.cur_len - 1] = ctx->sfn.buf[ctx->sfn.cur_len - 1] = '/';
      ctx->uri.buf[ctx->uri.cur_len]     = ctx->sfn.buf[ctx->sfn.cur_len]     = '\0';
    }

    /* Set "entry-related" fields */
    reusable.uri        = ctx->uri.buf;
    reusable.collection = S_ISDIR(entry->stat.st_mode);

    reusable.info->stat = *entry;

     /* Attach or descend */
    if (!ctx->walk_resource.resource->collection) {
      /* Add file */
      err = ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_MEMBER);
      if (err)
        break;  
    }
    else {
      /* Recursive into child */
      if ((err = dav_ns_walker(ctx, depth - 1)) != NULL)
        break;
    }

    /* Restore length, so next dav_buffer_place_mem overrites current entry */
    ctx->uri.cur_len  = uri_prev_len;
    ctx->sfn.cur_len  = path_prev_len;
  }

  /* Release resources */
  if (dmlite_closedir(this_res->info->ctx, dir))
    return dav_shared_new_error(this_res->info->request, this_res->info->ctx, 0,
                             "Could not close the directory %s", ctx->sfn.buf);

  if (err)
    return err;

  apr_pool_destroy(subpool);

  /* Recover */
  ctx->walk_resource.resource    = this_res;
  ctx->uri.buf[ctx->uri.cur_len] = '\0';
  ctx->sfn.buf[ctx->sfn.cur_len] = '\0';

  /* Postfix iteration */
  if (ctx->params->walk_type & DAV_WALKTYPE_POSTFIX) {
    return ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_POSTFIX);
  }

  /* There we go */
  return NULL;
}

/**
 * Walk through the file system
 * @param params    Walk parameters
 * @param depth     In depth?
 * @param response  The response
 * @return          NULL on success
 */
static dav_error *dav_ns_walk(const dav_walk_params *params, int depth,
                               dav_response **response)
{
  dav_ns_walker_context  ctx = {0};
  dav_error              *err;

  ctx.params = params;

  ctx.walk_resource.walk_ctx = params->walk_ctx;
  ctx.walk_resource.pool     = params->pool;
  ctx.walk_resource.resource = params->root;

  dav_buffer_init(params->pool, &ctx.sfn, params->root->info->sfn);
  dav_buffer_init(params->pool, &ctx.uri, params->root->uri);

  err = dav_ns_walker(&ctx, depth);
  
  *response = ctx.walk_resource.response;

  return err;
}

/**
 * Get an ETag
 * @param resource The resource from where retrieve the tag
 * @return         The tag value
 */
static const char *dav_ns_getetag(const dav_resource *resource)
{
  return apr_psprintf(resource->pool, "%lx-%lx",
                      (unsigned long)resource->info->stat.stat.st_ino,
                      (unsigned long)resource->info->stat.stat.st_mtime);
}

/** Hooks */
const dav_hooks_repository dav_ns_hooks_repository =
{
    1,
    dav_ns_get_resource,
    dav_ns_get_parent,
    dav_ns_is_same,
    dav_ns_is_parent,
    dav_ns_open_stream,
    dav_ns_close_stream,
    dav_ns_write_stream,
    dav_ns_seek_stream,
    dav_ns_set_headers,
    dav_ns_deliver,
    dav_ns_create_collection,
    dav_ns_copy,
    dav_ns_move,
    dav_ns_remove,
    dav_ns_walk,
    dav_ns_getetag
};
