/**
 * @file   security.c
 * @brief  Security related functions. Implementation.
 * @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
 */
#include <apr_strings.h>
#include <dmlite/c/dmlite.h>
#include <httpd.h>
#include <http_log.h>
#include "security.h"

/**
 * Un-escape the user DN
 * @param dn The user DN
 * @return dn
 */
static char *unescape_dn(char *dn)
{
  /* ap_unescape_url does not replace + */
  char *p;
  for(p = dn; *p != '\0'; ++p) {
    if (*p == '+') *p = ' ';
  }
  
  ap_unescape_url(dn);
  return dn;
}



dmlite_credentials *dav_shared_get_user_credentials(request_rec *r,
                                                    dmlite_context  *ctx,
                                                    const char  *anon_usr, const char *anon_grp,
                                                    apr_array_header_t* trusted)
{
  dmlite_credentials  *info;
  int                  i;
  const char          *s = NULL;
  char                *p = NULL;

  /* Recover if present in the notes */
  info = (dmlite_credentials*)apr_table_get(r->connection->notes,
                                           "dav_shared_user_info");
  if (info)
    return info;

  /* Create user information
   * We use the connection pool here because it has to be persistent through
   * requests */
  info = apr_pcalloc(r->connection->pool, sizeof(dmlite_credentials));
  info->fqans = apr_pcalloc(r->connection->pool, sizeof(char*) * DAV_SHARED_MAX_FQANS);

  /* Client IP */
#if AP_SERVER_MAJORVERSION_NUMBER == 2 &&\
    AP_SERVER_MINORVERSION_NUMBER < 4
  info->remote_address = r->connection->remote_ip;
#else
  info->remote_address = r->useragent_ip;
#endif

  /* Get user information from the connection notes */
  for (i = 0; (s = apr_table_get(r->connection->notes,
                                 apr_psprintf(r->pool, "GRST_CRED_AURI_%d", i)));
                                 ++i) {
    if (strncmp(s, "dn:", 3) == 0 && !info->client_name && (p = index(s, '/'))) {
      info->client_name = unescape_dn(apr_pstrdup(r->connection->pool, p));
    }
    if (strncmp(s, "fqan:", 5) == 0 && info->nfqans < DAV_SHARED_MAX_FQANS &&
        (p = index(s, '/'))) {
      info->fqans[info->nfqans++] = unescape_dn(apr_pstrdup(r->connection->pool, p));
    }
  }

  /* If not found, try with request notes */
  if (!i) {
    for (i = 0; (s = apr_table_get(r->notes, apr_psprintf(r->pool, "GRST_CRED_AURI_%d", i))); ++i) {
      if (strncmp(s, "dn:", 3) == 0 && !info->client_name && (p = index(s, '/'))) {
        info->client_name = unescape_dn(apr_pstrdup(r->connection->pool, p));
      }
      if (strncmp(s, "fqan:", 5) == 0 && info->nfqans < DAV_SHARED_MAX_FQANS && (p = index(s, '/'))) {
        info->fqans[info->nfqans++] = unescape_dn(apr_pstrdup(r->connection->pool, p));
      }
    }
  }

  /* Again, if not found, try with mod_ssl directly (won't work with proxies) */
  if (!i) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "mod_gridsite didn't give us anything. Trying with mod_ssl!");

    /* Since we cache this struct per connection, need to duplicate the string using
       the connection pool, or we may have a "use after free" */
    info->client_name = apr_pstrdup(r->connection->pool,
                                    apr_table_get(r->subprocess_env, "SSL_CLIENT_S_DN"));
    if (info->client_name != NULL) {
      i = 1;
    }
  }

  /* i = 0, that means there is no user certificate */
  if (!i) {
    /* Try to fallback to the anon, but keeping DN */
    if (anon_usr && anon_grp) {
      info->client_name = apr_pstrdup(r->connection->pool, anon_usr);
      info->nfqans = 1;
      info->fqans = apr_pcalloc(r->connection->pool, sizeof(const char*));
      info->fqans[0] = apr_pstrdup(r->connection->pool, anon_grp);

      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "No certificate presented. Falling back to %s:%s",
                    anon_usr, anon_grp);

      info->mech = "NONE";
    }
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "No certificate presented.");
      errno = EACCES;
      return NULL;
    }
  }
  else {
    info->mech = "X509";
  }
  
  /* Maybe the DN could be a trusted one */
  if (trusted != NULL) {
    char  is_trusted;
    for (i = 0, is_trusted = 0; i < trusted->nelts && !is_trusted; ++i) {
      const char *dn;

      dn = ((const char**)trusted->elts)[i];
      is_trusted = (strcmp(dn, info->client_name) == 0);
    }

    if (is_trusted) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Found trusted DN (%s)", info->client_name);
      /* We trust this DN, so try and check if it is presenting a DN header */
      const char *dn_header = apr_table_get(r->headers_in, "X-Auth-Dn");
      if (dn_header != NULL) {
        const char *fqan, *trusted_dn;

        trusted_dn = info->client_name;
        info->client_name = apr_pstrdup(r->connection->pool, dn_header);
        info->nfqans    = 0;

        /* Try fqans */
        for (i = 0;
             (fqan = apr_table_get(r->headers_in, apr_psprintf(r->pool, "X-Auth-Fqan%d", i))) != NULL;
             ++i) {
          info->fqans[info->nfqans++] = apr_pstrdup(r->pool, fqan);
        }

        ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
                      "Trusted DN (%s) acting on behalf of %s",
                      trusted_dn, info->client_name);
      }
      else {
        ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
                      "Trusted DN (%s) acting as itself!!",
                      info->client_name);
      }
    }
  }

  /* Save for later (keep-alive connections) */
  apr_table_setn(r->connection->notes, "dav_shared_user_info", (const char*)info); /* I am cheating here a bit :) */
 
  /* Done here */
  return info;
}
